/* WALKminsat */
/* this code  is just a minsat code as oppose to walkminsat.c that is used by walkQBF*/
/* Maintained by Henry Kautz <kautz@cs.washington.edu> */
/* Last Change by Gabriel Corvera 08/23/2005 23:12 print_sol created to show solution in the verifier format*/
/*  last changes Sept 1st, Yannet  */
#define VERSION "walkminsat-v1"

/************************************/
/* Compilation flags                */
/************************************/

/* If the constant DYNAMIC is set, then dynamic arrays are used instead of static arrays.
   This allows very large or very small problems to be handled, without
   having to adjust the constants MAXATOM and/or MAXCLAUSE.  However, on some
   architectures the compiled program is about 25% slower, because not
   as many optimizations can be performed. */

/* #define DYNAMIC 1 */

/***************************************************************/
/* Please select only one of following flags: UNIX, ANSI or NT */
/* Description:                                                */
/*    UNIX:  uses some non-POSIX unix time routines, compile   */
/*           successful under SunOS 5.6, time accuracy 1/60sec */
/*    ANIS:  use POSIX time routines, can compile under all    */
/*           unix and NT, time accuracy is 1sec                */ 
/*    NT:    use NT/Win32 multimedia routines, compile under   */
/*           NT only, time accuracy is 1ms                     */
/*           Uses NT "timeGetTime" function:                   */
/*             Header: Decl. in Mmsystem.h; include Windows.h. */
/*             Library: Use Winmm.lib.                         */
/***************************************************************/

#define UNIX 1
/* #define ANSI 1 */
/* #define NT 1 */


/* Define BIGINT to be the type used for the "cutoff" variable.
   Under gcc "long long int" gives a 64 bit integer.
   Program will still function using a 32-bit value, but it
   limit size of cutoffs that can be specified. */

#ifdef UNIX
#define BIGINT long long int
#endif
#ifdef NT
#define BIGINT __int64
#endif
#ifdef ANSI
#define BIGINT long int
#endif


/************************************/
/* Standard includes                */
/************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <sys/types.h>
#include <limits.h>
#include <signal.h>
#include <assert.h>

#ifdef UNIX
  #include <sys/times.h>
  #include <sys/time.h>
#endif
#ifdef NT
  #include <time.h>
  #include <windows.h>
  #include <mmsystem.h>
#endif
#ifdef ANSI
  #include <sys/time.h>
#endif

#ifndef UNIX   /* ANSI or NT */
  #define random() rand()
  #define srandom(seed) srand(seed)
#endif

/* Sometimes CLK_TCK is defined in a weird place */
#ifndef CLK_TCK
  #define CLK_TCK 60
#endif

#define GETMINLIT(x) x
#define GETLITMIN(x) x


/************************************/
/* Constant parameters              */
/************************************/


#define MAXATOM 150000		/* maximum possible number of atoms */

#define STOREBLOCK 2000000	/* size of block to malloc each time */
#define MAXCLAUSE 500000	/* maximum possible number of clauses */

#define TRUE 1
#define FALSE 0

#define MAXLENGTH 500           /* maximum number of literals which can be in any clause */

/************************************/
/* Internal constants               */
/************************************/

enum heuristics { RANDOM, BEST, WORST, BALANCE, BESTMIN,GS};

#define NOVALUE -1


#define Var(CLAUSE, POSITION) (ABS(clause[CLAUSE][POSITION]))

static int scratch;
#define ABS(x) ((scratch=(x))>0?(scratch):(-scratch))

#define BIG 100000000

/************************************/
/* Main data structures             */
/************************************/

/* Atoms start at 1 */
/* Not a is recorded as -1 * a */
/* One dimensional arrays are statically allocated. */
/* Two dimensional arrays are dynamically allocated in */
/* the second dimension only.  */

int numatom;
int numclause;
int numliterals;

/**
 * COMPLETING WALKMINSAT
 * 
 */
int complete = 0;
int min_solution[MAXATOM+1];
int banned[MAXCLAUSE];
int used[MAXATOM+1];
int last=0;
int worknumoccurence[2*MAXATOM+1];
void update_ocurrences(int index,int value);
void undo_ocurrences(int index,int value);
void search_minsat();
void search_minsat_util(int,int);
int choose_branch();

/**
 * Definition of Data Structures and Variables Needed to solve
 * MinSAT
 */

int * clause[MAXCLAUSE];	/* clauses to be satisfied */
/* indexed as clause[clause_num][literal_num] */
int size[MAXCLAUSE];		/* length of each clause */
int false[MAXCLAUSE];		/* clauses which are false */
int lowfalse[MAXCLAUSE];
int wherefalse[MAXCLAUSE];	/* where each clause is listed in false */
int numtruelit[MAXCLAUSE];	/* number of true literals in each clause */

int *occurence[2*MAXATOM+1];	/* where each literal occurs */
  /* indexed as occurence[literal+MAXATOM][occurence_num] */

int numoccurence[2*MAXATOM+1];	/* number of times each literal occurs */
	
	
int atom[MAXATOM+1];		/* value of each atom */ 
int bestminsat;			/* value used to know the best minsat number of true clauses */
int wheretrue[MAXCLAUSE]; /* where each clause is listed in true */
int true[MAXCLAUSE]; /* clauses which are true */	
int whereonetrue[MAXCLAUSE]; /* where each cluase is listed in onetrue */
int onetrue[MAXCLAUSE]; /* clauses which have only on true literal */
int numtrue;			/* number of true clauses */
int numonetrue; 		/* number of only one true literal clauses */	

int lowatom[MAXATOM+1];
int solution[MAXATOM+1];

int changed[MAXATOM+1];		/* step at which atom was last flipped */

int breakcount[MAXATOM+1];	/* number of clauses that become unsat if var if flipped */
int makecount[MAXATOM+1];	/* number of clauses that become sat if var if flipped */
int D_balance[MAXATOM+1];         /* change in balance if var if flipped */


int TheTrueOne[MAXCLAUSE];

int numfalse;			/* number of false clauses */


/************************************/
/* Global flags and parameters      */
/************************************/

int status_flag = 0;		/* value returned from main procedure */
int abort_flag;
int make_flag;
int balance_flag;

int heuristic = WORST;

int numerator = NOVALUE;	/* make random flip with numerator/denominator frequency */
int denominator = 100;		
int tabu_length;		/* length of tabu list */


BIGINT numflip;                /* number of changes so far*/
int numrun = 10;
BIGINT cutoff = 100000;


int numtry = 0;			/* total attempts at solutions */


/* Initialization options */

char initfile[100] = { 0 };
int initoptions = FALSE;

/* Randomization */

unsigned int seed;  /* Sometimes defined as an unsigned long int */

#ifdef UNIX
   struct timeval tv;
   struct timezone tzp;
#endif
#ifdef NT
   DWORD win_time;     /* elapsed time in ms, since windows boot up */
#endif


/************************************/
/* Forward declarations             */
/************************************/

void parse_parameters(int argc,char *argv[]);
void print_parameters(int argc, char * argv[]);
void init_minsat(void);

int pickrandom(void);
int pickbest(void);
int pickworst(void);
int pickbalance(void);
int pickbestmin(void);
int pickgs(void);

char * heuristic_names[] = { "random", "best","worst","balance","bestmin","gs"};

static int (*pickcode[])(void) =
      {pickrandom, pickbest, pickworst,pickbalance,pickbestmin,pickgs};


double elapsed_seconds(void);
int countunsat(void);
void scanone(int argc, char *argv[], int i, int *varptr);
void scanonell(int argc, char *argv[], int i, BIGINT *varptr);
void init(char initfile[], int initoptions);
void initprob(void); 
void flipatom(int toflip);

void print_false_clauses(long int lowbad);
void save_false_clauses(long int lowbad);
void print_low_assign(long int lowbad);
void save_low_assign(void);
void save_solution(void);

void print_current_assign(void);
void handle_interrupt(int sig);

long super(int i);

void print_sol_file(char * filename);
void print_sol_cnf(void);
void print_sol(void);

/************************************/
/* Main                             */
/************************************/


int main(int argc,char *argv[])
{

#ifdef UNIX
  gettimeofday(&tv,&tzp);
  seed = (unsigned int)((( tv.tv_sec & 0177 ) * 1000000) + tv.tv_usec);
#endif
#ifdef NT
  seed = (unsigned int)(timeGetTime());
#endif
#ifdef ANSI
  seed = (unsigned int)(time());
#endif
  parse_parameters(argc, argv);
  srandom(seed);
  //print_parameters(argc, argv);
  initprob();
  //signal(SIGINT, handle_interrupt);
  //    abort_flag = FALSE;
  //(void) elapsed_seconds();
  
  while (! abort_flag && numtry < numrun) {
	    
    numtry++;
    init(initfile, initoptions);
    numflip = 0;
		  
    while(numflip < cutoff) {	
      numflip++;
      flipatom((pickcode[heuristic])());
      
      if(numtrue<bestminsat){
	save_solution();
	bestminsat = numtrue;
      }	    		
      
    }
  }
  //expertime = elapsed_seconds();
  
  if(complete){
	  printf("Minsat: %d\n",bestminsat);	
	  search_minsat();
  }
  
  printf("\nMinimal Solution Found: (%d true clauses):\n",bestminsat);
  print_sol();
		    
  return status_flag;
}


void parse_parameters(int argc,char *argv[])
{
    int i;
    int temp;

    for (i=1;i < argc;i++)
    {
		if (strcmp(argv[i],"-seed") == 0)
		    scanone(argc,argv,++i,&seed);
		//else if (strcmp(argv[i],"-out") == 0 && i<argc-1)
		//  strcpy(outfile, argv[++i]);
		else if (strcmp(argv[i],"-status") == 0)
		    status_flag = 1;
		else if (strcmp(argv[i],"-cutoff") == 0)
		    scanonell(argc,argv,++i,&cutoff);
		else if (strcmp(argv[i],"-random") == 0)
		    heuristic = RANDOM;
		else if (strcmp(argv[i],"-best") == 0)
		    heuristic = BEST;
		else if (strcmp(argv[i],"-worst") == 0)
		    heuristic = WORST;
		else if (strcmp(argv[i],"-bestmin") == 0)
		    heuristic = BESTMIN;		    	
		else if (strcmp(argv[i],"-gs") == 0)
		    heuristic = GS;		    	    
		else if (strcmp(argv[i],"-complete") == 0)
		    complete = 1;
		else if (strcmp(argv[i],"-noise") == 0){
		    scanone(argc,argv,++i,&numerator);
		    if (i < argc-1 && sscanf(argv[i+1],"%i",&temp)==1){
			denominator = temp;
			i++;
		    }
		}
	        
		else if (strcmp(argv[i],"-tries") == 0 || strcmp(argv[i],"-restart") == 0)
		    scanone(argc,argv,++i,&numrun);
		
		else 
		{
		    fprintf(stderr, "General parameters:\n");
		    fprintf(stderr, "  -seed N -cutoff N -restart N\n");
		    fprintf(stderr, "Heuristics:\n");
		    fprintf(stderr, "  -random -worst -gs -bestmin\n");
		    fprintf(stderr, "  -noise N or -noise N M (default M = 100)\n");
  		    fprintf(stderr, "Complete Search:\n");
  		    fprintf(stderr, "  -complete\n");
		    exit(-1);
		}
	}
    //base_cutoff = cutoff;
    
    if (numerator==NOVALUE){
	switch(heuristic) {
	case WORST:
	case BEST:
	  numerator = 50;
	    break;
	default:
	    numerator = 0;
	    break;
	}
    }
}



void print_parameters(int argc, char * argv[])
{
  int i;

  printf("%s\n", VERSION);
  printf("command line =");
  for (i=0;i < argc;i++){
    printf(" %s", argv[i]);
  }
  printf("\n");
  printf("seed = %u\n",seed);
  printf("cutoff = %Li\n",cutoff);
  printf("tries = %i\n",numrun);

  printf("heuristic = ");
  switch(heuristic)
    {
    default:
      printf("%s", heuristic_names[heuristic]);
      break;
    }
  if (numerator>0){
    printf(", noise %d / %d", numerator, denominator);
  }
  
  printf("\n");
}



//void print_statistics_final(void)
//{
//  seconds_per_flip = expertime / totalflip;
//  printf("\ntotal elapsed seconds = %f\n", expertime);
//  printf("average flips per second = %d\n", (long)(totalflip/expertime));
//}



void scanone(int argc, char *argv[], int i, int *varptr)
{
    if (i>=argc || sscanf(argv[i],"%i",varptr)!=1){
	fprintf(stderr, "Bad argument %s\n", i<argc ? argv[i] : argv[argc-1]);
	exit(-1);
    }
}

void scanoneu(int argc, char *argv[], int i, unsigned int *varptr)
{
    if (i>=argc || sscanf(argv[i],"%u",varptr)!=1){
	fprintf(stderr, "Bad argument %s\n", i<argc ? argv[i] : argv[argc-1]);
	exit(-1);
    }
}

void scanonell(int argc, char *argv[], int i, BIGINT *varptr)
{
    if (i>=argc || sscanf(argv[i],"%Li",varptr)!=1){
	fprintf(stderr, "Bad argument %s\n", i<argc ? argv[i] : argv[argc-1]);
	exit(-1);
    }
}




void init(char initfile[], int initoptions)
{
  int i;
  int j;
  int thetruelit;

  for(i = 0;i < numclause;i++){
    numtruelit[i] = 0;
    banned[i]=0;
  }
  numtrue = 0;
  numonetrue = 0;
  
  /* Random Initialization of Atom */
  for(i = 1;i < numatom+1;i++){
    atom[i] = random()%2;
    breakcount[i] = 0;
    makecount[i] = 0;
    D_balance[i] =0;
  }
  

  /* Initialize breakcount and makecount in the following: */
  for(i = 0;i < numclause;i++)
    {
      for(j = 0;j < size[i];j++)
	{
	  if((clause[i][j] > 0) == atom[ABS(clause[i][j])]){ 
	    numtruelit[i]++;
	    thetruelit = clause[i][j];
	  }
	}
      if(numtruelit[i] == 0)
	{
	  for(j = 0;j < size[i];j++){
	    makecount[ABS(clause[i][j])]++;
	  }
	}
      else{
	if(numtruelit[i] > 0)
	  {
	    true[numtrue] = i;
	    wheretrue[i] = numtrue;
	    numtrue++;
	    //change this
	    TheTrueOne[i]=thetruelit;  //to assign a true literal even to non onetrue clauses
	  }
	if (numtruelit[i] == 1)
	  {
	    //TheTrueOne[i]=thetruelit;
	    breakcount[ABS(thetruelit)]++;
	    onetrue[numonetrue] = i;
	    whereonetrue[i] = numonetrue;
	    numonetrue++; 
	  }
      }
    }
}
	
	void initprob(void)
	{
	    int i;
	    int j;
	    int lastc;
	    int nextc;
	    int *storeptr;
	    int freestore;
	    int lit;
	
	    while ((lastc = getchar()) == 'c')
	      {
		  while ((nextc = getchar()) != EOF && nextc != '\n');
	      }
	    ungetc(lastc,stdin);
	    if (scanf("p cnf %i %i",&numatom,&numclause) != 2)
	      {
		  fprintf(stderr,"Bad input file\n");
		  exit(-1);
	      }
	    if(numatom > MAXATOM)
	      {
		  fprintf(stderr,"ERROR - too many atoms\n");
		  exit(-1);
	      }
	      
	    bestminsat = numclause;
	      
	
	    if(numclause > MAXCLAUSE)                     
	      {                                      
		  fprintf(stderr,"ERROR - too many clauses\n"); 
		  exit(-1);                              
	      }                                        
	    freestore = 0;
	    numliterals = 0;
	    for(i = 0;i < 2*MAXATOM+1;i++){
	      numoccurence[i] = 0;
	      worknumoccurence[i]=0;
	    }
	    for(i = 0;i < numclause;i++)
	      {
		  size[i] = -1;
		  if (freestore < MAXLENGTH)
		    {
			storeptr = (int *) malloc( sizeof(int) * STOREBLOCK );
			freestore = STOREBLOCK;
			fprintf(stderr,"allocating memory...\n");
		    }
		  clause[i] = storeptr;
		  do
		    {
			size[i]++;
			if(size[i] > MAXLENGTH)
			  {
			      printf("ERROR - clause too long\n");
			      exit(-1);
			  }
			if (scanf("%i ",&lit) != 1)
			  {
			      fprintf(stderr, "Bad input file\n");
			      exit(-1);
			  }
			if(lit != 0)
			  {
			      *(storeptr++) = lit; /* clause[i][size[i]] = j; */
			      freestore--;
			      numliterals++;
			      numoccurence[lit+MAXATOM]++;
			      
			  }
		    }
		  while(lit != 0);
	      }
	    if(size[0] == 0)
	      {
		  fprintf(stderr,"ERROR - incorrect problem format or extraneous characters\n");
		  exit(-1);
	      }
	
	    for(i = 0;i < 2*MAXATOM+1;i++)
	      {
		  if (freestore < numoccurence[i])
		    {
			storeptr = (int *) malloc( sizeof(int) * STOREBLOCK );
			freestore = STOREBLOCK;
			fprintf(stderr,"allocating memory...\n");
		    }
		  occurence[i] = storeptr;
		  freestore -= numoccurence[i];
		  storeptr += numoccurence[i];
		  numoccurence[i] = 0;
	      }
	
	    for(i = 0;i < numclause;i++)
	      {
		  for(j = 0;j < size[i];j++)
		    {
			occurence[clause[i][j]+MAXATOM]
			  [numoccurence[clause[i][j]+MAXATOM]] = i;
			numoccurence[clause[i][j]+MAXATOM]++;
			worknumoccurence[clause[i][j]+MAXATOM]++;
		    }
	      }
	}


void flipatom(int toflip)
{
  int i,j;		
  int toenforce;		
  register int cli;
  register int lit;
  int numocc;
  register int sz;
  register int * litptr;
  int * occptr;

  if(atom[toflip] > 0){
    toenforce = -toflip;
  }
  else{
    toenforce = toflip;
  }
  
  /* flipping the atom */
  atom[toflip] = 1-atom[toflip];
  
  numocc = numoccurence[MAXATOM+toenforce];
  occptr = occurence[MAXATOM+toenforce];
  
  for(i = 0; i < numocc; i++)
    {
      /* cli = occurence[MAXATOM+toenforce][i]; */
      cli = *(occptr++);
      
      if (++numtruelit[cli] == 1){
	/* add to one true and true since now it is a true clause */
	TheTrueOne[cli]=toenforce;
	onetrue[numonetrue] = cli;
	whereonetrue[cli] = numonetrue;
	numonetrue++;
	true[numtrue] = cli;
	wheretrue[cli] = numtrue;
		      /* Increment toflip's breakcount */
	      breakcount[toflip]++;
	numtrue++;
	/* Decrement the makecount of all vars in the clause */
	sz = size[cli];
	litptr = clause[cli];
	for (j=0; j<sz; j++){
	  /* lit = clause[cli][j]; */
	  lit = *(litptr++);
	  makecount[ABS(lit)]--;
	}
      }
      else if (numtruelit[cli] == 2){
	/* remove from one true */   
	numonetrue--;
	onetrue[whereonetrue[cli]]=onetrue[numonetrue]; 
	whereonetrue[onetrue[numonetrue]]=whereonetrue[cli];
		      sz = size[cli];
	      litptr = clause[cli];
	      for (j=0; j<sz; j++){
		  /* lit = clause[cli][j]; */
		  lit = *(litptr++);
		  if( ((lit > 0) == atom[ABS(lit)]) &&
		     (toflip != ABS(lit)) ){
		      breakcount[ABS(lit)]--;
		      break;
		  }
	      }
      }
    }
  
  numocc = numoccurence[MAXATOM-toenforce];
  occptr = occurence[MAXATOM-toenforce];
  
  for(i = 0; i < numocc ;i++)
    {
      /* cli = occurence[MAXATOM-toenforce][i]; */
      cli = *(occptr++);
      
      if (--numtruelit[cli] == 0){
	/* It is now a false clause, erase it from one and onetrue */
	numonetrue--;
	numtrue--;
	      /* Decrement toflip's breakcount */
	      breakcount[toflip]--;	
	true[wheretrue[cli]] = true[numtrue];
	wheretrue[true[numtrue]] = wheretrue[cli];
	onetrue[whereonetrue[cli]] = onetrue[numonetrue];
	whereonetrue[onetrue[numonetrue]] = whereonetrue[cli];		  
	/* Increment the makecount of all vars in the clause */
	sz = size[cli];
	litptr = clause[cli];
	for (j=0; j<sz; j++){
	  /* lit = clause[cli][j]; */
	  lit = *(litptr++);
	  makecount[ABS(lit)]++;
	}
      }
      else if (numtruelit[cli] == 1){
	/* now it is a one true clause */
	sz = size[cli];
	litptr = clause[cli];
	for (j=0; j<sz; j++){
	  lit = *(litptr++);
	  if ((lit > 0) == atom[ABS(lit)]){
	    TheTrueOne[cli]=lit;
	    breakcount[ABS(lit)]++;
	    break;
	  }
	}
	
	onetrue[numonetrue] = cli;
	whereonetrue[cli] = numonetrue;
	numonetrue++;   
      }
    }
  
  
}
	
/***************Heuristics********************/

int pickbalance()
{

  int i,number=10,nummake;
  int tofix[MAXLENGTH];
  int best[MAXLENGTH];
  register int numbest;
  register int bestvalue;
  register int var;

  if(numonetrue==0){
    return pickrandom();
  }

  for (i=0;i<number;i++)
    tofix[i] = onetrue[random()%numonetrue];
  for (i=0;i<number;i++){
    var = ABS(TheTrueOne[tofix[i]]);
    nummake = makecount[var];
    if (nummake<=bestvalue){
      if (nummake<bestvalue)
	numbest=0;
      bestvalue = nummake;
      best[numbest++] = var;
    }
  }
  if (bestvalue>0 && (random()%denominator < numerator))
    return ABS(TheTrueOne[tofix[0]]);
  
  if (numbest == 1) 
    return best[0];
  return best[random()%numbest];
  

}



/* choose a random clauses in onetrue and flip the sat lit */
int pickrandom(void){

  int tofix;
  
  if(numonetrue>0){
    tofix = onetrue[random()%numonetrue];			
  }else if(numtrue>0){
    tofix = true[random()%numtrue];	
  }else{
    assert(0); /* must not enter here */
  }

  //change this for the case numonetrue=0
  return ABS(TheTrueOne[tofix]);
}



/* choose form two random clauses in onetrue the literal that satisies less clauses  */
int pickworst(void){

  int i,number=10,nummake;
  int tofix[MAXLENGTH];
  int best[MAXLENGTH];
  register int numbest;
  register int bestvalue;
  register int var;

  numbest = 0;
  bestvalue = BIG;
  
  if (numtrue==0){
    printf("ERROR in pickworst \n");
    return -1;
  }

  if(numonetrue>0){
    for (i=0;i<number;i++)
      tofix[i] = onetrue[random()%numonetrue];
  }
  else if(numtrue>0){
//    printf(" numonetrue=0 \n");
    for (i=0;i<number;i++)
      tofix[i] = true[random()%numtrue];	
  }

  //change this ... for the case numonetrue=0
  for (i=0;i<number;i++){
    var = ABS(TheTrueOne[tofix[i]]);
    nummake = makecount[var];
    if (nummake<=bestvalue){
      if (nummake<bestvalue)
	numbest=0;
      bestvalue = nummake;
      best[numbest++] = var;
    }
  }
  if (bestvalue>0 && (random()%denominator < numerator))
    return ABS(TheTrueOne[tofix[0]]);
  
  if (numbest == 1) 
    return best[0];
  return best[random()%numbest];
}

int pickbestmin(void){

  int i,number=10,numbreak;
  int tofix[MAXLENGTH];
  int best[MAXLENGTH];
  register int numbest;
  register int bestvalue;
  register int var;

  numbest = 0;
  bestvalue = 0;
  
  if (numtrue==0){
    printf("ERROR in pickbestmin \n");
    return -1;
  }

  if(numonetrue>0){
    for (i=0;i<number;i++)
      tofix[i] = onetrue[random()%numonetrue];
  }
  else if(numtrue>0){
    printf(" numonetrue=0 \n");
    for (i=0;i<number;i++)
      tofix[i] = true[random()%numtrue];	
  }

  //change this ... for the case numonetrue=0
  for (i=0;i<number;i++){
    var = ABS(TheTrueOne[tofix[i]]);
    numbreak = breakcount[var];
    if (numbreak>=bestvalue){
      if (numbreak>bestvalue)
		numbest=0;
      bestvalue = numbreak;
      best[numbest++] = var;
    }
  }
  if (bestvalue>0 && (random()%denominator < numerator))
    return ABS(TheTrueOne[tofix[0]]);
  
  if (numbest == 1) 
    return best[0];
  return best[random()%numbest];
}

int pickgs(void){

  int i,number=10,gs;
  int tofix[MAXLENGTH];
  int best[MAXLENGTH];
  register int numbest;
  register int bestvalue;
  register int var;

  numbest = 0;
  bestvalue = -BIG;
  
  if (numtrue==0){
    printf("ERROR in pickgs \n");
    return -1;
  }

  if(numonetrue>0){
    for (i=0;i<number;i++)
      tofix[i] = onetrue[random()%numonetrue];
  }
  else if(numtrue>0){
    printf(" numonetrue=0 \n");
    for (i=0;i<number;i++)
      tofix[i] = true[random()%numtrue];	
  }

  //change this ... for the case numonetrue=0
  for (i=0;i<number;i++){
    var = ABS(TheTrueOne[tofix[i]]);
    gs = breakcount[var]-makecount[var];
    if (gs>=bestvalue){
      if (gs>bestvalue)
		numbest=0;
      bestvalue = gs;
      best[numbest++] = var;
    }
  }
  if (bestvalue>0 && (random()%denominator < numerator))
    return ABS(TheTrueOne[tofix[0]]);
  
  if (numbest == 1) 
    return best[0];
  return best[random()%numbest];
}

int pickbest(void)
{
  int numbreak;
    int tofix;
    int clausesize;
    int i;		
    int best[MAXLENGTH];
    register int numbest;
    register int bestvalue;
    register int var;

    tofix = false[random()%numfalse];
    clausesize = size[tofix];
    numbest = 0;
    bestvalue = BIG;

    for (i=0; i< clausesize; i++){
      var = ABS(clause[tofix][i]);
      numbreak = breakcount[var];
      if (numbreak<=bestvalue){
			if (numbreak<bestvalue) numbest=0;
			bestvalue = numbreak;
			best[numbest++] = var;
      }
    }

    if (bestvalue>0 && (random()%denominator < numerator))
      return ABS(clause[tofix][random()%clausesize]);

    if (numbest == 1) return best[0];
    return best[random()%numbest];
}



int countunsat(void)
{
  int i, j, unsat, bad, lit, sign;
  
  unsat = 0;
  for (i=0;i < numclause;i++)
    {
      bad = TRUE;
      for (j=0; j < size[i]; j++)
	{
	  lit = clause[i][j];
	  sign = lit > 0 ? 1 : 0;
	  if ( atom[ABS(lit)] == sign )
	    {
	      bad = FALSE;
	      break;
	    }
	}
      if (bad)
	unsat++;
    }
  return unsat;
}


double elapsed_seconds(void) 
{ 
	double answer;

#ifdef ANSI
	static long prev_time = 0;
	time( &long_time );
	/* Note:  time(&t) returns t in seconds, so do not need to /CLK_TCK */
	answer = long_time - prev_time;
	prev_time = long_time;
#endif
#ifdef NT
	static DWORD prev_time = 0;
	win_time = timeGetTime();
	/* Note:  return value of timeGetTime() is ms, so divide by 1000*/
	answer = (double)(win_time - prev_time) / (double)1000; 
	prev_time = win_time;
#endif
#ifdef UNIX
    static struct tms prog_tms;
    static long prev_times = 0;
    (void) times(&prog_tms);
    answer = ((double)(((long)prog_tms.tms_utime)-prev_times))/((double) CLK_TCK);
    prev_times = (long) prog_tms.tms_utime;
#endif
	return answer; 
}




void print_sol_cnf(void)
{
    int i;
    for(i = 1;i < numatom+1;i++)
	printf("v %i\n", solution[i] == 1 ? i : -i);
}


void print_sol(void)
{
  int i;
  for(i = 1;i < numatom+1;i++){
    printf("%i ", solution[i]);		
  }
  printf("\n");
}

void print_sol_file(char * filename)
{
    FILE * fp;
    int i;

    if ((fp = fopen(filename, "w"))==NULL){
	fprintf(stderr, "Cannot open output file\n");
	exit(-1);
    }
    for(i = 1;i < numatom+1;i++){
	fprintf(fp, " %i", solution[i] == 1 ? i : -i);
	if (i % 10 == 0) fprintf(fp, "\n");
    }
    if ((i-1) % 10 != 0) fprintf(fp, "\n");
    fclose(fp);
}



void print_current_assign(void)
{
    int i;

    printf("Begin assign at flip = %d \n", numflip);
    for (i=1; i<=numatom; i++){
	printf(" %d", atom[i]==0 ? -i : i);
	if (i % 10 == 0) printf("\n");
    }
    if ((i-1) % 10 != 0) printf("\n");
    printf("End assign\n");
}


	void
	save_solution(void)
	{
	    int i;
	
	    for (i=1; i<=numatom; i++)
	      solution[i] = atom[i];
	}


void init_minsat(){


  heuristic = WORST;

  
  status_flag = 0;		/* value returned from main procedure */



  numerator = 50;	/* make random flip with numerator/denominator frequency */
  denominator = 100;		

  numrun = 1; //I change this  august 26, Yannet
  numtry=0;
  cutoff = 5000; //100000
  
	
}


int choose_branch(){
	int i,choosen=0,maxocc=-1;
	for(i=MAXATOM-numatom;i<MAXATOM+numatom+1;i++){		
		if(i==MAXATOM || used[(i<MAXATOM?MAXATOM-i:i-MAXATOM)]!=0){			
			continue;
		}
		if(maxocc<worknumoccurence[i]){
			maxocc=worknumoccurence[i];
			choosen=(i<MAXATOM?MAXATOM-i:i-MAXATOM);
		}			
	}
	
	assert(choosen>0);
	used[choosen]=1;
	return choosen;
}		



void undo_ocurrences(int index,int value){
	int numocc, *occptr, cli,i,j,sz,lit,*litptr;
	int toenforce = (value?index:-index);
	
	numocc = numoccurence[MAXATOM+toenforce];
  	occptr = occurence[MAXATOM+toenforce];
  
  	for(i = 0; i < numocc ;i++)
    {
      /* cli = occurence[MAXATOM-toenforce][i]; */
      cli = *(occptr++);	
	  sz = size[cli];
	  litptr = clause[cli];

	  if(banned[cli]==1){
		  for (j=0; j<sz; j++){
		  	/* lit = clause[cli][j]; */
		  	lit = *(litptr++);
		  	worknumoccurence[MAXATOM+lit]++;
		  }
	  }
	  banned[cli]--;
      
      
    }
}

void update_ocurrences(int index,int value){
	int numocc, *occptr, cli,i,j,sz,lit,*litptr;
	int toenforce = (value?index:-index);
	
	numocc = numoccurence[MAXATOM+toenforce];
  	occptr = occurence[MAXATOM+toenforce];
  
  	for(i = 0; i < numocc ;i++)
    {
      /* cli = occurence[MAXATOM-toenforce][i]; */
      cli = *(occptr++);	
	  sz = size[cli];
	  litptr = clause[cli];
	  
	  if(banned[cli]==0){
		  for (j=0; j<sz; j++){
		  	/* lit = clause[cli][j]; */
		  	lit = *(litptr++);
		  	worknumoccurence[MAXATOM+lit]--;
		  }
	  }
	  banned[cli]++;
      
    }
}

void search_minsat(){
	int vars = 0;
	int i,numocc1=0,numocc2=0;
	
	for(i=0;i<MAXATOM+1;i++){
		min_solution[i] = -1;
		used[i]=0;
	}

	i = choose_branch();

	min_solution[i]=0;
	numocc1=worknumoccurence[MAXATOM-i];
	update_ocurrences(i,0);
	search_minsat_util(numocc1, vars+1);
	undo_ocurrences(i,0);
	min_solution[i]=-1;
	
	min_solution[i]=1;
	numocc2=worknumoccurence[MAXATOM+i];
	update_ocurrences(i,1);
	search_minsat_util(numocc2, vars+1);
	undo_ocurrences(i,1);
	min_solution[i]=-1;
		
}

void search_minsat_util(int sum, int vars){	
	int numocc1,numocc2,i;

	if(vars == (numatom)){		
		if(sum<bestminsat){
			//end of tree keep solution			
			bestminsat=sum;
			printf("BESTMINSAT: %d\n",bestminsat);
			for(i=1;i<numatom+1;i++){
				solution[i] = min_solution[i];
			}			
		}
		return;
	}else if(sum >= bestminsat){
		//printf("Depth: %d %d\n",vars,sum);
		return; //We are done, just cut the tree	
	}else{
		i = choose_branch();
		
		numocc1=worknumoccurence[MAXATOM-i];
		min_solution[i]=0;
		update_ocurrences(i,0);
		search_minsat_util(sum+numocc1, vars+1);
		undo_ocurrences(i,0);
		min_solution[i]=-1;
		
		numocc2=worknumoccurence[MAXATOM+i];
		min_solution[i]=1;
		update_ocurrences(i,1);
		search_minsat_util(sum+numocc2, vars+1);
		undo_ocurrences(i,1);
		min_solution[i]=-1;
		
		used[i]=0;
	}
	return;
}


